"""Data model technology interface"""
from abc import ABC
from opsorchestrator.core.decorator.class_decorators import classproperty

class Technology(ABC):
    @classproperty
    def name(self) -> str:
        """Each subclass must provide a name"""
        return self.__name__



"""
Execution unit interface
with Execution Unit Execution parameters and Validation Result
"""
from abc import ABC, abstractmethod
from opsorchestrator.core.data_model import DataModel
from opsorchestrator.core.decorator.class_decorators import classproperty

class ExecutionUnit(ABC):
    """
    Execution Unit Interface
    """
    _operation_id = None
    _user_session_id = None
    _operation_name = None

    @classproperty
    def name(cls) -> str:
        """The name of an execution unit"""
        raise NotImplementedError("Name was never defined")

    @classproperty
    def data_model(cls) -> DataModel:
        raise NotImplementedError("Subclasses must define .data_model")

    @classproperty
    def unique_name(self) -> str:
        return self.__name__ if isinstance(self,type) else self.__class__.__name__


    @classmethod
    def emit(cls, message):
        """Emit a message"""
        print(message)


    @classmethod
    @abstractmethod
    def execute(
        cls,
        operation_name : str,
        operation_id : str,
        user_session_id : str,
        required_data : dict
        ): #to be done; Arguments has to be passed from upper layer not accessed via sub_operation or operation
        """Should provide concrete implementation for execution"""
        raise NotImplementedError("Execution unit has to implement execute method")


"""General file to define custom decorators within classes"""
from functools import update_wrapper
class classproperty:
    """Decorator to allow treating a property as classproperty mimcing classmethod"""
    def __init__(self, fget):
        self.fget = fget
        update_wrapper(self, fget)
    def __get__(self, _, owner=None):
        return self.fget(owner)



from opsorchestrator.core.data_model.technology import Technology
from opsorchestrator.core.data_model.component import Component

class Datalake(Technology):
    pass

class ThreatFeed(Component):
    pass


"""Contracts for Zscaler Technology"""
from opsorchestrator.core.data_model.technology import Technology
from opsorchestrator.core.data_model.component import Component

class Zscaler(Technology):
    pass
class ZIA(Component):
    pass
class ZPA(Component):
    pass
class ZCC(Component):
    pass



from opsorchestrator.core.data_model.technology import Technology
from opsorchestrator.core.data_model.component import Component


class Microsoft(Technology):
    pass


class Mde(Component):
    pass


# pylint: disable=no-self-argument
"""
Operation Interface
"""
from abc import ABC, abstractmethod
from opsorchestrator.core.decorator.class_decorators import classproperty
from opsorchestrator.core.data_model import DataModel

class Operation(ABC):
    """Abstract base class defining the interface for an Operation."""
    @classproperty
    @abstractmethod
    def name(cls) -> str:
        raise NotImplementedError("Subclasses must define .name")

    @classproperty
    @abstractmethod
    def title(cls) -> str:
        raise NotImplementedError("Subclasses must define .title")

    @classproperty
    @abstractmethod
    def description(cls) -> str:
        raise NotImplementedError("Subclasses must define .description")

    @classproperty
    @abstractmethod
    def timeout(cls) -> int:
        raise NotImplementedError("Subclasses must define .timeout")

    @classproperty
    @abstractmethod
    def result_expiration_period(cls) -> int:
        raise NotImplementedError("Subclasses must define .result_expiration_period")

    @classproperty
    @abstractmethod
    def data_model(self) -> DataModel:
        raise NotImplementedError("Subclasses must define .data_model")



"""Contracts for ServiceNow Technology"""
from opsorchestrator.core.data_model.technology import Technology
from opsorchestrator.core.data_model.component import Component

class ServiceNow(Technology):
    pass

class Incident(Component):
    pass



from opsorchestrator.core.data_model.technology import Technology
from opsorchestrator.core.data_model.component import Component

class Network(Technology):
    pass
class TestConnectivity(Component):
    pass


"""Contracts for OpsOrchestrator"""
from opsorchestrator.core.data_model.technology import Technology
from opsorchestrator.core.data_model.component import Component

class OpsOrchestrator(Technology):
    pass
class Frontend(Component):
    pass


"""
This module contains the RBAC handling for Operations
"""
def role(value: str):
    """
    Decorator that attaches predefined roles to a class
    and adds a can_access(token) method that uses them.
    """
    def decorator(cls):
        return cls
    return decorator



"""Data model element"""
import re
from opsorchestrator.core.data_source import DataSourceScope
from opsorchestrator.core.data_model.component import Component
from opsorchestrator.core.data_model.technology import Technology

class DataElement():
    """Data element interface to define a key within a component"""
    def __init__(self, name : str, technology : Technology, component : Component, element_type : type, \
                  is_auth : bool = False, is_optional : bool = False, \
                  scope : DataSourceScope = DataSourceScope.OPERATION, \
                  expiration_period = None, 
                  skip_execution_if_not_found = False): #Justification: Already imported via TYPE_CHECKING, pylint: disable=used-before-assignment
        super().__setattr__("_name", name)
        self._element_type = element_type
        self._is_auth = is_auth
        self._scope = scope
        self.technology = technology
        self.component = component
        self._is_optional = is_optional
        self._expiration_period = expiration_period
        self._skip_execution_if_not_found = skip_execution_if_not_found

    def __setattr__(self, name, value):
        if name == "_technology":
            raise AttributeError("Direct assignment to _technology is forbidden. Use technology only.")
        if name == "_component":
            raise AttributeError("Direct assignment to _component is forbidden. Use component only.")
        if name == "_name":
            raise AttributeError("Direct assignment to _name is forbidden. Use DataElement initializer only.")
        super().__setattr__(name, value)

    @property
    def element_type(self):
        return self._element_type

    @property
    def is_authentication_element(self):
        return self._is_auth

    @property
    def is_optional(self):
        return self._is_optional

    @property
    def skip_execution(self):
        """Skip execution of that execution unit if data is absent."""
        return self._skip_execution_if_not_found

    @property
    def technology(self):
        return self._technology #Justification: element is set directly using __setattr__ pylint: disable=no-member

    @technology.setter
    def technology(self, _technology):
        if  not issubclass(_technology, Technology):
            raise TypeError("Technology has to be of type Technology")
        super().__setattr__("_technology", _technology)

    @property
    def component(self):
        return self._component #Justification: element is set directly using __setattr__ pylint: disable=no-member

    @component.setter
    def component(self, _component):
        if  not issubclass(_component, Component):
            raise TypeError("Component has to be of type Component")
        super().__setattr__("_component", _component)

    @property
    def name(self):
        return self._name #Justification: element is set directly using __setattr__ pylint: disable=no-member

    @name.setter
    def name(self,name):
        """
        Validates that the provided `name` string adheres to specific naming rules:
        - Must be a string.
        - Must be lowercase.
        - Can only contain lowercase letters and underscores.

        Args:
            name (str): The name string to validate.

        Raises:
            TypeError: If `name` is not a string.
            ValueError: If `name` is not lowercase or contains invalid characters.

        Returns:
            bool: True if validation passes.
        """
        # Allowed: lowercase letters, underscore
        name_re = re.compile(r"^[a-z_]+$")
        if not isinstance(name, str):
            raise TypeError(f"name: {name} must be a string")
        if name != name.lower():
            raise ValueError(f"name: {name} must be lowercase")
        if not name_re.fullmatch(name):
            raise ValueError(f"name: {name} may only contain lowercase letters, and underscore")
        super().__setattr__("_name", name)

    @property
    def fully_qualified_name(self):
        return f"{self.technology.__name__.lower()}.{self.component.__name__.lower()}.{self.name}"

    @property
    def unique_name(self):
        return self.fully_qualified_name.replace(".","_")

    @property
    def scope(self) -> DataSourceScope:
        return self._scope

    @scope.setter
    def scope(self, scope):
        self._scope = scope

    @property
    def expiration_period(self):
        return self._expiration_period



"""Data model that defines the data fully qualified name"""
from abc import ABC
from typing import List
from opsorchestrator.core.data_model.data_element import DataElement

class DataModel(ABC):
    """Abstract base class for DataModel"""
    def __init__(self, required_data, \
                 output_data, preconditions = None):
        if isinstance(required_data, list):
            for data_element in required_data:
                if not isinstance(data_element, DataElement):
                    raise TypeError("Required data has to contain data elements")
        else:
            raise TypeError("Output data has to be a list")

        if isinstance(output_data, list):
            for data_element in output_data:
                if not isinstance(data_element, DataElement):
                    raise TypeError("Output data has to contain data elements")
        else:
            raise TypeError("Output data has to be a list")

        self._required_data_list = required_data
        self._output_data_list = output_data
        self._preconditions = preconditions if preconditions else []

    @property
    def required_data(self) -> List[DataElement]:
        return self._required_data_list
    @property
    def output_data(self) -> List[DataElement]:
        return self._output_data_list

    def is_valid(self, operation_id, user_session_id) -> bool:
        return all([condtion.check(operation_id, user_session_id) \
                    for condtion in self._preconditions]) \
                        if self._preconditions else True